# -*- coding: utf-8 -*-
# NSKinetics: simulation of Non-Steady state enzyme Kinetics and inhibitory phenomena
# Copyright (C) 2025-, Sarang S. Bhagwat <sarangbhagwat.developer@gmail.com>
# 
# This module is under the MIT open-source license. See 
# https://github.com/sarangbhagwat/nskinetics/blob/main/LICENSE
# for license details.

import nskinetics as nsk
import numpy as np
from warnings import filterwarnings

def test_simple_ESP_inverse_modeling_and_doe(plot=True, 
                                             show_progress=False,
                                             show_output=True,
                                             show_warnings=True):
    
    if not show_warnings: filterwarnings("ignore")
    
    np.random.seed(3221)
    
    #%% Create a SpeciesSystem object with defined kinetic parameters
    sp_sys = nsk.SpeciesSystem('sp_sys', 
                           ['E', 'S', 'ES', 'P'], # enzyme, substrate, enzyme-substrate complex, product
                           concentrations=[1e-4, 1e-4, 0., 0.])
    
    # Describe reactions by writing chemical equations and kinetic parameter info
    reactions = [
                'E + S <-> ES; kf = 12.0, kb = 10.0', # kf = kon, kb = koff
                'ES -> E + P; kf = 32.0' # kf = kcat (enzyme turnover number)
                ]
    
    # Generate a ReactionSystem from strings
    rxn_sys = nsk.ReactionSystem(ID='rxn_sys', 
                                     reactions=reactions,
                                     species_system=sp_sys)
    
    KM = (rxn_sys.reactions[1].kf + rxn_sys.reactions[0].kb)/rxn_sys.reactions[0].kf
    
    #%% Define batches to run
    
    batch_concs = [
                   # np.array([1e-4, 1e-4, 0., 0.]),
                   # np.array([0.1, 0.89656207, 0., 0.]),
                   np.array([0.08966552, 2.5, 0., 0.]),
                   # np.array([0.08966552, 0.31041379, 0., 0.]),
                   ]
    batch_t_spans = [
                     # [0, 0.5*24*3600],
                     # [0, 300],
                     [0, 300],
                     # [0, 300],
                     ]
    
    n_batches = len(batch_concs)
    batch_indices = range(n_batches)
    
    #%% Simulate all batches and save results
    
    dfs_results = []
    
    for i, concentrations, t_span in zip(batch_indices, 
                                         batch_concs, 
                                         batch_t_spans):
        sp_sys.concentrations = concentrations
        # Simulate the ReactionSystem
        rxn_sys.solve(t_span=t_span,
                      sp_conc_for_events={'S':1e-6},
                      # filename=f'solution{i}',
                      save_events_df=True,)                              
        dfs_results.append(rxn_sys._solution_dfs[0])
        # Plot results
        if plot:
            rxn_sys.plot_solution()
            rxn_sys.plot_solution(sps_to_include=['ES'])
    
    #%% Fit kinetic parameters to saved results from simulating with original set of kinetic parameters
    
    # print(rxn_sys._timeout_solve_ivp)
    rxn_sys.fit_reaction_kinetic_parameters_to_data(
                                                    # data=[f'solution{i}.xlsx'
                                                    #       for i in batch_indices],
                                                    data=dfs_results,
                                                    p0=1*np.ones(len(rxn_sys.reaction_kinetic_params)),
                                                    use_only=[
                                                              # 'E', 
                                                              'S', 
                                                              # 'P'
                                                              ],
                                                    options={'disp':show_progress},
                                                    # plot_during_fit=True,
                                                    show_progress=show_progress,
                                                    show_output=show_output,
                                                    n_minimize_runs=2,
                                                    n_de_runs=3,
                                                    differential_evolution_kwargs={'maxiter':100,
                                                                                   'bounds': [(0,5e4) for i in range(len(rxn_sys.reaction_kinetic_params))],
                                                                                   'polish': False,
                                                                                   'disp': show_progress,
                                                                                   'workers': 1,
                                                                                   'popsize': 10,
                                                                                   'tol': 1e-3,
                                                                                   }
                                                    )
    
    #%% Simulate batches with inverse-modeled rxn_sys
    for i, concentrations, t_span in zip(range(len(batch_concs)), 
                                         batch_concs, 
                                         batch_t_spans):
        sp_sys.concentrations = concentrations
        # Simulate the ReactionSystem
        rxn_sys.solve(t_span=t_span,
                      sp_conc_for_events={'S':1e-6},
                      # filename=f'solution{i}',
                      save_events_df=True,
                      )                  
        # Plot results            
        if plot:
            rxn_sys.plot_solution()
            rxn_sys.plot_solution(sps_to_include=['ES'])
        
    #%% Design of experiments
    param_keys = rxn_sys.reaction_kinetic_param_keys
    candidate_initials = {'E': np.linspace(1e-4, 0.1, 30), 'S': np.linspace(1e-4, 1., 30), 'ES': [0.0], 'P': [0.0]}
    spike_options = None
    t_eval = np.linspace(0, 30, 50)
    output_idx = [sp_sys.index('S'), 
                  sp_sys.index('P')]  # Measuring S and P
    
    best_expts = nsk.doe.optimal_design(
        rxn_sys,
        param_keys,
        candidate_initials,
        t_eval,
        spike_options=spike_options,
        output_idx=output_idx,
        epsilon=1e-4,
        top_n=3,
        show_output=show_output,
        )
    
    # Tests
    
    assert np.allclose(rxn_sys._fitsol[1], 
                       1.,
                       rtol=1e-3, atol=1e-5)
    
    return rxn_sys 

def test_simple_ESP_inverse_modeling_and_doe_with_known_kcat_KM(plot=True, 
                                             show_progress=False,
                                             show_output=True,
                                             show_warnings=True):
    
    if not show_warnings: filterwarnings("ignore")
    
    np.random.seed(3221)
    
    #%% Create a SpeciesSystem object with defined kinetic parameters
    sp_sys = nsk.SpeciesSystem('sp_sys', 
                           ['E', 'S', 'ES', 'P'], # enzyme, substrate, enzyme-substrate complex, product
                           concentrations=[1e-4, 1e-4, 0., 0.])
    
    # Describe reactions by writing chemical equations and kinetic parameter info
    reactions = [
                'E + S <-> ES; kf = 12.0, kb = 10.0', # kf = kon, kb = koff
                'ES -> E + P; kf = 32.0' # kf = kcat (enzyme turnover number)
                ]
    
    # Generate a ReactionSystem from strings
    rxn_sys = nsk.ReactionSystem(ID='rxn_sys', 
                                     reactions=reactions,
                                     species_system=sp_sys)
    
    KM = (rxn_sys.reactions[1].kf + rxn_sys.reactions[0].kb)/rxn_sys.reactions[0].kf
    
    #%% Define batches to run
    
    batch_concs = [
                   # np.array([1e-4, 1e-4, 0., 0.]),
                   np.array([0.08966552, 0.58624828, 0., 0.]),
                   # np.array([0.08966552, 0.62072759, 0., 0.]),
                   # np.array([0.0621069, 0.10353793, 0., 0.]),
                   ]
    batch_t_spans = [
                     # [0, 0.5*24*3600],
                     [0, 300],
                     # [0, 30],
                     # [0, 30],
                     ]
    
    n_batches = len(batch_concs)
    batch_indices = range(n_batches)
    
    #%% Simulate all batches and save results
    
    dfs_results = []
    
    for i, concentrations, t_span in zip(batch_indices, 
                                         batch_concs, 
                                         batch_t_spans):
        sp_sys.concentrations = concentrations
        # Simulate the ReactionSystem
        rxn_sys.solve(t_span=t_span,
                      sp_conc_for_events={'S':1e-6},
                      # filename=f'solution{i}',
                      save_events_df=True,)                              
        dfs_results.append(rxn_sys._solution_dfs[0])
        # Plot results
        if plot:
            rxn_sys.plot_solution()
            rxn_sys.plot_solution(sps_to_include=['ES'])
    
    #%% Fit kinetic parameters to saved results from simulating with original set of kinetic parameters
    
    rxn_sys.reactions[1]._freeze_kf = True
    rxn_sys.reactions[0]._freeze_kb = True
    
    def set_koff(p, kcat=rxn_sys.reactions[1].kf, KM=KM):
        # With a known KM and known (frozen) kcat, 
        # set koff based on the kon value being used by the solver
        kon = p[0]
        rxn_sys.reactions[0]._kb = KM*kon - kcat
    
    
    rxn_sys.fit_reaction_kinetic_parameters_to_data(
                                                    # data=[f'solution{i}.xlsx'
                                                    #       for i in batch_indices],
                                                    data=dfs_results,
                                                    p0=1*np.ones(len(rxn_sys.reaction_kinetic_params)),
                                                    use_only=[
                                                              # 'E', 
                                                              'S', 
                                                              # 'P'
                                                              ],
                                                    call_before_each_solve=[set_koff],
                                                    options={'disp':show_progress},
                                                    # plot_during_fit=True,
                                                    show_progress=show_progress,
                                                    show_output=show_output,
                                                    n_minimize_runs=2,
                                                    n_de_runs=3,
                                                    differential_evolution_kwargs={'maxiter':100,
                                                                                   'bounds': [(0,5e4) for i in range(len(rxn_sys.reaction_kinetic_params))],
                                                                                   'polish': False,
                                                                                   'disp': show_progress,
                                                                                   'workers': 1,
                                                                                   'popsize': 10,
                                                                                   'tol': 1e-3,
                                                                                   }
                                                    )
    
    #%% Simulate batches with inverse-modeled rxn_sys
    for i, concentrations, t_span in zip(range(len(batch_concs)), 
                                         batch_concs, 
                                         batch_t_spans):
        sp_sys.concentrations = concentrations
        # Simulate the ReactionSystem
        rxn_sys.solve(t_span=t_span,
                      sp_conc_for_events={'S':1e-6},
                      # filename=f'solution{i}',
                      save_events_df=True,
                      )                  
        # Plot results            
        if plot:
            rxn_sys.plot_solution()
            rxn_sys.plot_solution(sps_to_include=['ES'])
        
    #%% Design of experiments
    param_keys = rxn_sys.reaction_kinetic_param_keys
    candidate_initials = {'E': np.linspace(1e-4, 0.1, 30), 'S': np.linspace(1e-4, 1., 30), 'ES': [0.0], 'P': [0.0]}
    spike_options = None
    t_eval = np.linspace(0, 30, 50)
    output_idx = [sp_sys.index('S'), 
                  sp_sys.index('P')]  # Measuring S and P
    
    best_expts = nsk.doe.optimal_design(
        rxn_sys,
        param_keys,
        candidate_initials,
        t_eval,
        spike_options=spike_options,
        output_idx=output_idx,
        epsilon=1e-4,
        top_n=3,
        show_output=show_output,
        )
    
    # Tests
    
    assert np.allclose(rxn_sys._fitsol[0], 
                       np.array([11.99998575]),
                       rtol=1e-3, atol=1e-8)
    
    assert np.allclose(rxn_sys._fitsol[1], 
                       0.9999999999992784,
                       rtol=1e-4, atol=1e-8)
    
    assert np.allclose(best_expts[0]['y0'],
                       np.array([0.08966552, 
                                 0.6552069, 
                                 0., 
                                 0.]),
                       rtol=1e-5, atol=1e-8)
    
    assert np.allclose(best_expts[0]['FIM'],
                       np.array([[0.0018309195499738703]]),
                       rtol=1e-1, atol=1e-8)
    
    assert np.allclose(best_expts[0]['score'],
                       0.0018309195499738703,
                       rtol=1e-1, atol=1e-8)

def test_michaelis_menten_steady_state_inverse_modeling(plot=False, 
                                                        show_progress=False,
                                                        show_output=False,
                                                        show_warnings=False):
    Vmax_orig = 100.
    KM_orig = 2.5
    S_data = np.linspace(0, 12, 50)
    V_data = nsk.michaelis_menten(S=S_data, Vmax=Vmax_orig, KM=KM_orig)
    
    E0 = 2.
    kcat_fit, KM_fit = nsk.fit_michaelis_menten(data=None, 
                                    E0=E0, 
                                    S_data=S_data, 
                                    V_data=V_data, 
                                    normalize=False)
    Vmax_fit = kcat_fit*E0
    
    assert (np.allclose([Vmax_fit, KM_fit], [Vmax_orig, KM_orig]))
    
    
    kcat_fit, KM_fit = nsk.fit_michaelis_menten(data=None, 
                                    E0=E0, 
                                    S_data=S_data, 
                                    V_data=V_data, 
                                    normalize=True)
    Vmax_fit = kcat_fit*E0
    
    assert (np.allclose([Vmax_fit, KM_fit], [Vmax_orig, KM_orig]))
    
    


# -*- coding: utf-8 -*-
# NSKinetics: simulation of Non-Steady state enzyme Kinetics and inhibitory phenomena
# Copyright (C) 2025-, Sarang S. Bhagwat <sarangbhagwat.developer@gmail.com>
# 
# This module is under the MIT open-source license. See 
# https://github.com/sarangbhagwat/nskinetics/blob/main/LICENSE
# for license details.

__version__ = '0.2.3'

# %% Initialize NSKinetics 

from . import species
from .species import *

from . import reactions
from .reactions import *

from . import reaction_systems 
from .reaction_systems import *

from . import doe
# from .doe import *

from . import utils

from . import tests

__all__ = (
     *species.__all__,
     *reactions.__all__,
     *reaction_systems.__all__,
     'utils',
     'doe',
     'tests',
     )



# -*- coding: utf-8 -*-
# NSKinetics: simulation of Non-Steady state enzyme Kinetics and inhibitory phenomena
# Copyright (C) 2025-, Sarang S. Bhagwat <sarangbhagwat.developer@gmail.com>
# 
# This module is under the MIT open-source license. See 
# https://github.com/sarangbhagwat/nskinetics/blob/main/LICENSE
# for license details.

import numpy as np

__all__ = ('Species', 'SpeciesSystem',
           'SimpleSpecies', 'ComplexSpecies')

#%% Species and species system

class Species():
    """
    Abstract class for a chemical species.
    
    Parameters
    ----------
    ID : str
        ID.
    MW: float or int
        Molar mass (molecular weight) of this species.
        
    """
    def __init__(self, ID, MW=1.):
        self._ID = ID
        self._MW = MW
    
    @property
    def ID(self): 
        return self._ID
    
    @property
    def MW(self): 
        return self._MW
    
class SimpleSpecies(Species):
    """
    Abstract class for a single chemical species that is not a complex
    of other chemical species participating in the expected species system.
    
    Parameters
    ----------
    ID : str
        ID.
    MW: float or int
        Molar mass (molecular weight) of this species.
        
    """
    def __init__(self, ID, MW=1.):
        Species.__init__(self, ID=ID, MW=MW)

class ComplexSpecies(Species):
    """
    Represents a complex formed by two or more constituent Species.
    Must provide exactly one combination of these parameters:
        - ID and simple_species_system; 
        - components
    Other parameters are optional.
    
    Parameters
    ----------

    ID : str, optional
        If provided, must be a string created by conjugating IDs of components 
        with '~'. For components with non-1 stoichiometries, ID must include these
        in the format f'number*species_ID' (as shown below), with or without parentheses
        (either is fine, although will be stored with parentheses). 
        Can be conjugated in any order, but will be stored in alphabetic order of 
        component Species IDs. Defaults to a string automatically
        created by this method from components.
        E.g., if components == {'E':1, 'S':1, 'I':2}, ID defaults to 'E~2*(I)~S'.
        
    components : dict, optional
        If provided, must be a dictionary mapping Species objects or Species IDs 
        to their stoichiometric coefficients in the complex.
        E.g., {speciesA: 1, speciesB: 2}.

    MW: float or int, optional
        Molar mass (molecular weight) of this species. Defaults to the stoichiometry-adjusted
        sum of its component species' molecular weights.
            
    simple_species_system: SimpleSpeciesSystem, optional
        Simple species system that includes all components of this complex species.
        
    Attributes
    ----------
    ID : str
        ID of the complex species.
    components : dict
        Species objects with their stoichiometric coefficients.
    MW : float
        Molecular weight of the complex, automatically computed from the sum of the components.
    """
    
    def __init__(self, ID=None, simple_species_system=None, components=None, MW=None):
        if ID is None and components is None:
            raise ValueError('\nEither ID or components must be provided, but both were None.')
        
        if ID is None:
            if isinstance(components, list):
                components = {i: 1 for i in components}
            
            for k, v in list(components.items()):
                if isinstance(k, str):
                    k_sp = simple_species_system.species(k)
                    components[k_sp] = v
                    del components[k]
                    
            self._components = components
            ID = self.ID_from_components(components)
            
        elif components is None:
            self._components = self.components_from_ID(ID)
        
        else:
            ValueError(f'\nOnly one of ID or components must be provided, but both were provided (ID={ID}, components={components}).')
        
        if MW is None:
            # calculate the MW as the sum of its parts
            MW = sum(species.MW * stoich for species, stoich in components.items())
        
        Species.__init__(self, ID=ID, MW=MW)
        
    def ID_from_components(self, components):
        components_sorted = sorted(components.items(), key=lambda i: i[0].ID)
        ID = ''
        for c in components_sorted:
            if c[1]==1:
                ID+='~'+c[0].ID
            else:
                ID+='~'+str(c[1])+'*('+c[0].ID+')'
        ID = ID[1:]
        return ID
    
class SpeciesSystem():
    """
    Abstract class for a system of chemical species.
    Note this class will, by convention,
    use "sp" to denote a single species and
    use "sps" to denote multiple species.
    
    Parameters
    ----------
    ID : str
        ID.
    all_sps : list
        List of Species objects or strings. 
        If the latter, new Species objects will be created.
    concentrations : list or np.ndarray, optional
        Concentrations of species, indexed the same way as
        all_sps. Defaults to zero-array of length equal to
        that of all_sps.
    
    """
    def __init__(self, ID, all_sps, concentrations=None):
        self.ID = ID
        
        processed_all_sps = []
        for i in all_sps:
            if isinstance(i, str):
                processed_all_sps.append(Species(ID=i))
            elif isinstance(i, Species):
                processed_all_sps.append(i)
            else:
                raise TypeError(f"\nProvided member of all_sps '{i}' must be of type str or Species.\n")
        self.all_sps = processed_all_sps
        
        if concentrations is None:
            concentrations = np.zeros(len(processed_all_sps))
        else:
            concentrations = np.array(concentrations)
        
        self._concentrations = concentrations
        
    def indices(self, some_sps):
        all_sps = self.all_sps
        indices = [all_sps.index(i) for i in some_sps]
        return indices
    
    def index(self, sp):
        if isinstance(sp, Species):
            return self.all_sps.index(sp)
        elif isinstance(sp, str):
            return self.index_from_ID(sp)
    
    def index_from_ID(self, sp_ID):
        index = 0
        all_sps = self.all_sps
        for i in all_sps:
            if i.ID==sp_ID:
                return index
            index += 1
    
    def contains(self, sp):
        if sp in self.all_sps + self.all_sp_IDs:
            return True
        else:
            return False
        
    @property
    def all_sp_IDs(self):
        all_sps = self.all_sps
        all_sp_IDs = []
        for sp in all_sps:
            if isinstance(sp, str):
                all_sp_IDs.append(sp)
            else:
                all_sp_IDs.append(sp.ID)
        return all_sp_IDs
    
    def species(self, ID_or_index):
        if isinstance(ID_or_index, int):
            return self.all_sps[ID_or_index]
        elif isinstance(ID_or_index, str):
            return self.all_sps[self.all_sp_IDs.index(ID_or_index)]
        else:
            raise ValueError(f'\nID_or_index must be int or str, but instead got {ID_or_index} of type {type(ID_or_index)}.')
    
    @property
    def concentrations(self):
        return self._concentrations
    
    @concentrations.setter
    def concentrations(self, concentrations):
        self._concentrations = concentrations
    


# -*- coding: utf-8 -*-
# NSKinetics: simulation of Non-Steady state enzyme Kinetics and inhibitory phenomena
# Copyright (C) 2025-, Sarang S. Bhagwat <sarangbhagwat.developer@gmail.com>
# 
# This module is under the MIT open-source license. See 
# https://github.com/sarangbhagwat/nskinetics/blob/main/LICENSE
# for license details.

from . import reaction_system_gui
from .reaction_system_gui import *

__all__ = (
     *reaction_system_gui.__all__,
     )



# -*- coding: utf-8 -*-
# NSKinetics: simulation of Non-Steady state enzyme Kinetics and inhibitory phenomena
# Copyright (C) 2025-, Sarang S. Bhagwat <sarangbhagwat.developer@gmail.com>
# 
# This module is under the MIT open-source license. See 
# https://github.com/sarangbhagwat/nskinetics/blob/main/LICENSE
# for license details.

__all__ = ('create_function', 
           'codify')

def codify(statement):
    statement = replace_apostrophes(statement)
    statement = replace_newline(statement)
    return statement

def replace_newline(statement):
    statement = statement.replace('\n', ';')
    return statement

def replace_apostrophes(statement):
    statement = statement.replace('’', "'").replace('‘', "'").replace('“', '"').replace('”', '"')
    return statement

def create_function(code, namespace):
    def wrapper_fn(statement):
        def f(t, concs):
            namespace['t'] = t
            namespace['concs'] = concs
            exec(codify(statement), namespace)
            return namespace['y']
        return f
    function = wrapper_fn(code)
    return function



# -*- coding: utf-8 -*-
# NSKinetics: simulation of Non-Steady state enzyme Kinetics and inhibitory phenomena
# Copyright (C) 2025-, Sarang S. Bhagwat <sarangbhagwat.developer@gmail.com>
# 
# This module is under the MIT open-source license. See 
# https://github.com/sarangbhagwat/nskinetics/blob/main/LICENSE
# for license details.

from . import _create_function
from ._create_function import *

from . import _argument_checks
from ._argument_checks import *

from .import _fit
from ._fit import *

__all__ = (
     *_create_function.__all__,
     *_argument_checks.__all__,
     *_fit.__all__,
     )



# -*- coding: utf-8 -*-
# NSKinetics: simulation of Non-Steady state enzyme Kinetics and inhibitory phenomena
# Copyright (C) 2025-, Sarang S. Bhagwat <sarangbhagwat.developer@gmail.com>
# 
# This module is under the MIT open-source license. See 
# https://github.com/sarangbhagwat/nskinetics/blob/main/LICENSE
# for license details.

import nskinetics as nsk
import numpy as np
from warnings import filterwarnings

def test_simple_ESP_non_log_transform(plot=True, 
                    show_progress=True,
                    show_output=True,
                    show_warnings=True):
    
    if not show_warnings: filterwarnings("ignore")
    
    # Create a SpeciesSystem object
    sp_sys = nsk.SpeciesSystem('sp_sys', 
                           ['E', 'S', 'ES', 'P'], # enzyme, substrate, enzyme-substrate complex, product
                           concentrations=[1e-4, 1e-4, 0., 0.])
    
    # Describe reactions by writing chemical equations and kinetic parameter info
    reactions = [
                'E + S <-> ES; kf = 12.0, kb = 10.0', # kf = kon, kb = koff
                'ES -> E + P; kf = 32.0' # kf = kcat (enzyme turnover number)
                ]
    
    # Generate a ReactionSystem from strings
    rxn_sys = nsk.ReactionSystem(ID='rxn_sys', 
                                     reactions=reactions,
                                     species_system=sp_sys)
    
    rxn_sys._log_transform_concs = False
    # Simulate the ReactionSystem
    rxn_sys.solve(t_span=[0, 2*24*3600], # I want to simulate the system over 2 days
                     sp_conc_for_events={'S':1e-6}, # In addition to a full simulation,
                     )                              # I want to know the time at which [S] drops to 1e-6
    
    # Plot results
    if plot:
        rxn_sys.plot_solution()
        
        rxn_sys.plot_solution(sps_to_include=['ES'])
    
    # Tests
    assert np.allclose(rxn_sys._solution['t_events'], 
                       np.array([42219.44616989]), 
                       rtol=1e-3, atol=1e-3)
    
    assert np.allclose(rxn_sys._solution['y_events'], 
                       np.array([[9.99998909e-05],
                                [1.00000000e-06],
                                [1.09091871e-10],
                                [9.89998909e-05]]), 
                       rtol=1e-3, atol=1e-8)
    
    assert np.allclose(rxn_sys._solution['y'][:,50],
                       np.array([9.99958946e-05, 
                                 3.76342167e-05, 
                                 4.10542307e-09, 
                                 6.23616779e-05]),
                       rtol=1e-5, atol=1e-8)
    
    return rxn_sys


# -*- coding: utf-8 -*-
# NSKinetics: simulation of Non-Steady state enzyme Kinetics and inhibitory phenomena
# Copyright (C) 2025-, Sarang S. Bhagwat <sarangbhagwat.developer@gmail.com>
# 
# This module is under the MIT open-source license. See 
# https://github.com/sarangbhagwat/nskinetics/blob/main/LICENSE
# for license details.

import numpy as np

__all__ = ('is_number', 
           'is_array_of_numbers',
           'is_list_of_strings')

def is_number(i):
    return isinstance(i,float)\
            or isinstance(i,int)\
            or isinstance(i, np.floating)\
            or isinstance(i, np.integer)
            
def is_array_of_numbers(i):
    return isinstance(i, np.ndarray) and\
        (np.issubdtype(i.dtype, np.floating) or
         np.issubdtype(i.dtype, np.integer))

def is_list_of_strings(i):
    for s in i:
        if not isinstance(s, str):
            return False
    return True



# -*- coding: utf-8 -*-
# NSKinetics: simulation of Non-Steady state enzyme Kinetics and inhibitory phenomena
# Copyright (C) 2025-, Sarang S. Bhagwat <sarangbhagwat.developer@gmail.com>
# 
# This module is under the MIT open-source license. See 
# https://github.com/sarangbhagwat/nskinetics/blob/main/LICENSE
# for license details.

import numpy as np
from numba import njit
from warnings import warn
from .utils import read_equation_str

__all__ = ('Reaction', 'Rxn', 'IrreversibleReaction', 'IrrevRxn',
           'ReversibleReaction', 'RevRxn',
           'ChemicalEquation')

#%% Abstract chemical equation class

class ChemicalEquation():
    """
    Represents a chemical equation with defined stoichiometry.
    
    This class stores and formats the stoichiometry of a reaction,
    and is used to represent the equation symbolically and in string form.
    It can optionally be paired with an external object that also holds stoichiometric information.
    
    Parameters
    ----------
    ID : str
        Unique identifier for the chemical equation.
    species_system : SpeciesSystem
        System containing all chemical species.
    stoichiometry : array-like, optional
        Stoichiometric coefficients of all species (negative for reactants, positive for products).
    paired_obj : object, optional
        An object containing its own 'stoichiometry' attribute that this equation can reference 
        instead of passing 'stoichiometry' as an initialization argument.
    
    """
    def __init__(self, ID, 
                 species_system, 
                 stoichiometry=None, 
                 paired_obj=None, # can pass an object that has a parameter 'stoichiometry'
                 ):
        self.ID = ID
        self.species_system = species_system
        self._stoichiometry = stoichiometry if stoichiometry is not None\
            else paired_obj.stoichiometry
        self.paired_obj = paired_obj
        
    def __str__(self):
        lhs = ''
        rhs = ''
        for chem, stoich in zip(self.species_system.all_sps, self.stoichiometry):
            
            if stoich<0: 
                if not lhs=='': lhs+= ' + '
                if not np.abs(stoich)==1.: lhs+= str(-stoich) + ' '
                lhs+= chem.ID
            elif stoich>0: 
                if not rhs=='': rhs+= ' + '
                if not np.abs(stoich)==1.: rhs+= str(stoich) + ' '
                rhs+= chem.ID
                
        return f'{self.ID}: ChemicalEquation(' + lhs + ' -> ' + rhs + ')'
    
    def __repr__(self):
        return self.__str__()
    
    @property
    def stoichiometry(self):
        """
        The stoichiometric coefficients for this chemical equation.
        
        Returns
        -------
        np.ndarray
            Array of stoichiometric values (negative for reactants, positive for products).
        
        """
        return self._stoichiometry
    @stoichiometry.setter
    def stoichiometry(self, new_stoichiometry):
        self._stoichiometry = new_stoichiometry
        if self.paired_obj is not None:
            if not self.paired_obj.stoichiometry==new_stoichiometry:
                warn(f'Replaced {self.ID} stoichiometry with {new_stoichiometry}, but this does not match the paired_obj stoichiometry: {self.paired_obj.stoichiometry}.',
                     RuntimeWarning)
    
    @classmethod
    def from_string(cls, ID, equation_str, species_system, paired_obj=None):
        """
        Create a ChemicalEquation object from a string representation of a reaction.
        
        Parameters
        ----------
        ID : str
            Identifier for the chemical equation.
        equation_str : str
            String representing the chemical equation, e.g., 
            "A + B -> C",
            "A + B <-> C",
            "A + B -> C; kf=20",
            "A + B <-> C; kf=20, kb=40.5",
            "A + B <-> C; kf=20, kb = 10.5; rate_expr = kf*A*B - kb*C",
        species_system : SpeciesSystem
            System containing all species involved in the reaction.
        paired_obj : object, optional
            Object that may contain a stoichiometry attribute.

        Returns
        -------
        ChemicalEquation
            Instantiated ChemicalEquation object.
        float
            Forward rate constant (kf) parsed from the string, if present.
        float
            Backward rate constant (kb) parsed from the string, if present.
        str
            Rate expression (rate) parsed from the string, if present.
            
        """
        
        stoichiometry, kf, kb, rate_expr = read_equation_str(equation_str, 
                                                   species_system)
        return cls(ID=ID, 
                 species_system=species_system,
                 stoichiometry=stoichiometry,
                 paired_obj=paired_obj),\
            kf, kb, rate_expr
        
#%% Abstract reaction class
class AbstractReaction():
    """
    Abstract base class to define the stoichiometry of a chemical reaction.
    
    This class does not define any kinetic behavior (i.e., no rate constants),
    but sets up the stoichiometry from either a ChemicalEquation, lists/dicts of
    reactants and products, or a raw stoichiometry array.
    
    One and only one of the following must be provided:
        - reactants and products (as dict or list)
        - chem_equation
        - stoichiometry
        
    Parameters
    ----------
    ID : str
        Identifier for the reaction.
    species_system : SpeciesSystem
        System containing the chemical species.
    reactants : list or dict, optional
        Reactant species. If dict, keys are species IDs and values are stoichiometric coefficients.
    products : list or dict, optional
        Product species. Same format as `reactants`.
    chem_equation : ChemicalEquation, optional
        Equation object defining reaction stoichiometry.
    stoichiometry : array-like, optional
        Array of stoichiometric coefficients.
        
    """
    def __init__(self, ID, 
                 species_system,
                 reactants=None, products=None,
                 chem_equation=None,
                 stoichiometry=None,
                 ):
        self.ID = ID
        self.reactants = reactants
        self.products = products
        self.chem_equation = chem_equation
        self.species_system = species_system
        
        if stoichiometry is None:
            if chem_equation is None:
                stoichiometry = []
                if isinstance(reactants, dict) and isinstance(products, dict):
                    reactant_keys = reactants.keys()
                    product_keys = products.keys()
                    for i in species_system.all_sps:
                        if i.ID in reactant_keys: 
                            stoichiometry.append(reactants[i])
                        elif i.ID in product_keys: 
                            stoichiometry.append(products[i])
                        else:
                            stoichiometry.append(0.)
                elif isinstance(reactants, list) and isinstance(products, list):
                    for i in species_system.all_sps:
                        if i.ID in reactants: 
                            stoichiometry.append(-1)
                        elif i.ID in products: 
                            stoichiometry.append(1)
                        else:
                            stoichiometry.append(0.)
                else:
                    raise TypeError('\nGiven reactants and products must both be either lists or dicts.\n')
                
                self.stoichiometry = np.array(stoichiometry)
            
            else:
                self.stoichiometry = chem_equation.stoichiometry
        else:
            self.stoichiometry = stoichiometry
        
        if self.chem_equation is None:
            self.chem_equation = ChemicalEquation(ID=ID+'_eqn', 
                                                  species_system=species_system,
                                                  stoichiometry=self.stoichiometry)
        
#%% Reaction class

@njit(cache=True)
def dconcs_dt_v0_3(
    kf, kb, species_concs_vector, rxn_stoichs, rl_exps,
    reactant_indices, product_indices
    ):
    n_species = species_concs_vector.size
    
    # Early exit if any reactant is exhausted
    for idx in reactant_indices:
        if species_concs_vector[idx] <= 1e-20:
            return np.zeros(n_species, dtype=species_concs_vector.dtype)
        
    # Compute forward rate
    forward = kf
    for idx in reactant_indices:
        conc = species_concs_vector[idx]
        exp  = rl_exps[idx]
        forward *= conc ** exp
        
    # Compute backward rate
    backward = kb
    for idx in product_indices:
        conc = species_concs_vector[idx]
        exp  = rl_exps[idx]
        backward *= conc ** exp
        
    change = forward - backward
    
    # Evaluate limiting reactant (to avoid negative concentrations)
    min_ratio = 1e20
    limiting_found = False
    
    for i in range(n_species):
        stoich = rxn_stoichs[i]
        projected = species_concs_vector[i] + change * stoich
        if projected < 0.0:
            if stoich != 0.0:
                ratio = species_concs_vector[i] / abs(stoich)
                if ratio < min_ratio:
                    min_ratio = ratio
                    limiting_found = True
                    
    if limiting_found:
        change = min_ratio
        
    # Compute final concentration rate of change
    result = np.empty(n_species, dtype=species_concs_vector.dtype)
    for i in range(n_species):
        result[i] = change * rxn_stoichs[i]
        
    return result

class Reaction(AbstractReaction):
    """
    A complete chemical reaction object that includes kinetic parameters and rate law information.

    This class extends `AbstractReaction` by adding support for reaction rate constants,
    rate law exponents, reversible/irreversible behavior, and dynamic concentration updates.
    Includes logic to compute the rate of change in concentrations.
    
    One and only one of the following must be provided to define stoichiometry:
        - reactants and products (as dict or list)
        - chem_equation
        - stoichiometry
        
    Parameters
    ----------
    ID : str
        Identifier for the reaction.
    species_system : SpeciesSystem
        System containing the chemical species.
    kf : float
        Forward reaction rate constant.
    kb : float, optional
        Backward reaction rate constant (defaults to 0 for irreversible reactions).
    reactants : list or dict, optional
        Reactant species. If dict, keys are species IDs and values are stoichiometric coefficients.
    products : list or dict, optional
        Product species. Same format as `reactants`.
    chem_equation : ChemicalEquation, optional
        Equation object defining reaction stoichiometry.
    stoichiometry : array-like, optional
        Stoichiometric coefficients for all species.
    exponents : array-like, optional
        Rate law exponents for each species (same length as stoichiometry).
    get_exponents_from_stoich : bool, optional
        If True, use the absolute value of stoichiometry as the rate law exponents.
    freeze_kf : bool, optional
        If True, prevents modification of the forward rate constant, kf.
    freeze_kb : bool, optional
        If True, prevents modification of the backward rate constant, kb.
            
    """
    def __init__(self, ID, 
                 species_system,
                 kf,
                 kb=0.,
                 reactants=None, products=None,
                 chem_equation=None,
                 stoichiometry=None,
                 exponents=None,
                 get_exponents_from_stoich=False,
                 freeze_kf=False,
                 freeze_kb=False,
                 ):
        AbstractReaction.__init__(self, ID, 
                     species_system,
                     reactants=reactants, products=products,
                     chem_equation=chem_equation,
                     stoichiometry=stoichiometry,)
        
        stoich=self.stoichiometry
        
        if kf is None:
            kf = 0.
        if kb is None:
            kb = 0.
            
        self._kf = float(kf)
        self._kb = float(kb)
        
        self._freeze_kf = freeze_kf
        self._freeze_kb = freeze_kb
        
        self.get_exponents_from_stoich = get_exponents_from_stoich
        
        if exponents is None:
            if get_exponents_from_stoich:
                self.exponents = np.abs(len(stoich))
            else:
                self.exponents = np.ones(len(stoich))
        else:
            self.exponents = exponents
        self.reactant_indices = np.where(stoich<0)[0]
        self.product_indices = np.where(stoich>0)[0]
        self._load_full_string()
    
    @property
    def kf(self):
        """
        The forward reaction rate constant, kf.
    
        Returns
        -------
        float
            Forward reaction rate constant, kf.
            
        """
        return self._kf
    @kf.setter
    def kf(self, new_kf):
        """
        Set the forward reaction rate constant, kf.
        
        If `freeze_kf` is True, the value is not updated and a warning is issued.
        
        Parameters
        ----------
        new_kf : float
            New value for the forward reaction rate constant, kf.
            
        """
        if not self._freeze_kf:
            self._kf = new_kf
        else:
            warn(f'kf for Reaction {self.ID} was not changed to {new_kf} as _freeze_kf was True.\n',
                 RuntimeWarning)
    
    @property
    def kb(self):
        """
        The backward reaction rate constant, kf.
    
        Returns
        -------
        float
            Forward reaction rate constant, kb.
            
        """
        return self._kb
    @kb.setter
    def kb(self, new_kb):
        """
        Set the backward reaction rate constant, kb.
        
        If `freeze_kb` is True, the value is not updated and a warning is issued.
        
        Parameters
        ----------
        new_kb : float
            New value for the backward reaction rate constant, kb.
            
        """
        if not self._freeze_kb:
            self._kb = new_kb
        else:
            warn(f'kb for Reaction {self.ID} was not changed to {new_kb} as _freeze_kb was True.\n',
                 RuntimeWarning)
            
    def get_dconcs_dt(self):
        """
        Calculate the net rate of change in species concentrations for this reaction,
        using the current kinetic parameters and species concentrations.
        If self.species_system.log_transformed is True, returns rate 
        of change of np.log(concentrations) instead.
        
        Returns
        -------
        ndarray
            Change in concentration; or 
            zeros if kf and kb are both zero; or
            change in log(concentrations) if self.log_transformed is True.
        
        """
        kf, kb = self.kf, self.kb
        if kf==kb==0:
            return np.zeros(shape=self.species_system.concentrations.shape)
        
        return dconcs_dt_v0_3(kf=kf, 
                             kb=kb,
                             species_concs_vector=self.species_system._concentrations, 
                             rxn_stoichs=self.stoichiometry,
                             rl_exps=self.exponents,
                             reactant_indices=self.reactant_indices,
                             product_indices=self.product_indices)
    
    def _load_full_string(self):
        """
        Internal method to format and cache string representations of 
        the reaction's left-hand side, right-hand side, and kinetic parameters.
        
        """
        lhs = ''
        rhs = ''
        arrow = '->' if self.kb==0. else '<->'
        for chem, stoich in zip(self.species_system.all_sps, self.stoichiometry):
            
            if stoich<0: 
                if not lhs=='': lhs+= ' + '
                if not np.abs(stoich)==1.: lhs+= str(-stoich) + ' '
                lhs+= chem.ID
            elif stoich>0: 
                if not rhs=='': rhs+= ' + '
                if not np.abs(stoich)==1.: rhs+= str(stoich) + ' '
                rhs+= chem.ID
        param_info = f'kf={self.kf}'
        if arrow=='<->':
            param_info += ', ' + f'kb={self.kb}'
        self._lhs_string = lhs
        self._arrow_string = arrow
        self._rhs_string = rhs
        self._param_info_string = param_info
        self._full_string = f'{self.ID}: Reaction(' + lhs + ' ' + arrow + ' ' + rhs + '; ' + param_info + ')'
    
    def get_equation_str(self):
        return self._lhs_string + ' ' + self._arrow_string + ' ' + self._rhs_string
        
    def __str__(self):
        self._load_full_string()
        return self._full_string
    
    def __repr__(self):
        return self.__str__()
    
    @classmethod
    def from_equation(cls, ID, chem_equation, species_system, 
                      kf=None, # overrides any parameter info in the chem_equation string
                      kb=None, # overrides any parameter info in the chem_equation string
                      exponents=None, get_exponents_from_stoich=None):
        """
        Create a Reaction object from a ChemicalEquation object or string.
        
        Parameters
        ----------
        ID : str
            Identifier for the Reaction.
        chem_equation : str or ChemicalEquation
            Reaction as a string or ChemicalEquation object.
            String represents the chemical equation and may or may not
            include kf and kb (if not included, must pass at least kf 
                               -- kb defaults to zero -- or
                               both kf and kb as arguments). 
            E.g., 
            "A + B -> C",
            "A + B <-> C",
            "A + B -> C, kf=20",
            "A + B <-> C, kf=20, kb=40.5".
        species_system : SpeciesSystem
            System containing the involved species.
        kf : float, optional
            Forward rate constant (overrides any value parsed from str chem_equation).
        kb : float, optional
            Backward rate constant (overrides any value parsed from str chem_equation).
        exponents : ndarray, optional
            Rate law exponents.
        get_exponents_from_stoich : bool, optional
            Whether to use absolute stoichiometry as rate law exponents.
            
        Returns
        -------
        Reaction
            Instantiated Reaction object.
            
        """
        kf_, kb_, rate_expr_ = None, None, None
        freeze_kb = False
        if isinstance(chem_equation, str):
            freeze_kb = '->' in chem_equation and not '<->' in chem_equation
            chem_equation, kf_, kb_, rate_expr_ = ChemicalEquation.from_string(ID=ID+'_eqn', 
                                                         equation_str=chem_equation,
                                                         species_system=species_system)
        if kf is None:
            kf = kf_
        if kb is None:
            kb = kb_
        
        return cls(ID=ID,
                species_system=species_system,
                chem_equation=chem_equation,
                kf=kf,
                kb=kb,
                exponents=exponents,
                freeze_kf=False,
                freeze_kb=freeze_kb,
                get_exponents_from_stoich=get_exponents_from_stoich,)
        
Rxn = IrreversibleReaction = ReversibleReaction = IrrevRxn = RevRxn = Reaction



# -*- coding: utf-8 -*-
# NSKinetics: simulation of Non-Steady state enzyme Kinetics and inhibitory phenomena
# Copyright (C) 2025-, Sarang S. Bhagwat <sarangbhagwat.developer@gmail.com>
# 
# This module is under the MIT open-source license. See 
# https://github.com/sarangbhagwat/nskinetics/blob/main/LICENSE
# for license details.

import os
import nskinetics

nskinetics_filepath = nskinetics.__file__

# def test_RxnSys_from_SBML(filename='michaelis_menten_valid.xml',
#                           plot=True, 
#                           show_progress=True,
#                           show_output=True,
#                           show_warnings=True):
#     delimiter = None
#     for i in ['/', '\\']:
#         if i in nskinetics_filepath:
#             delimiter=i
    
#     SBML_file_location = nskinetics_filepath.replace('__init__.py', '')
#     SBML_file_location += 'tests'
#     # os.chdir(nskinetics_filepath+SBML_file_location)
#     rxn_sys = nskinetics.ReactionSystem.from_SBML(filepath=SBML_file_location+delimiter+filename)
    
#     return rxn_sys


# -*- coding: utf-8 -*-
# NSKinetics: simulation of Non-Steady state enzyme Kinetics and inhibitory phenomena
# Copyright (C) 2025-, Sarang S. Bhagwat <sarangbhagwat.developer@gmail.com>
# 
# This module is under the MIT open-source license. See 
# https://github.com/sarangbhagwat/nskinetics/blob/main/LICENSE
# for license details.

from sklearn.metrics import r2_score
from scipy.optimize import minimize
from scipy.optimize import differential_evolution
import numpy as np

__all__ = ('fit_multiple_dependent_variables',)

def fit_multiple_dependent_variables(f, 
                                     xdata, ydata,
                                     p0,
                                     r2_score_multioutput='uniform_average',
                                     n_minimize_runs=2,
                                     n_de_runs=5,
                                     minimize_kwargs=None,
                                     differential_evolution_kwargs=None,
                                     show_progress=False,
                                     **kwargs):
    """
    Fit a model function to multiple dependent variables using a shared set of parameters.
    
    Parameters
    ----------
    f : callable
        A model function that takes two arguments: an array of independent variables `xdata`
        and a parameter array `p`, and returns an array of shape (N, M) or (M, N), where N is
        the number of dependent variables and M is the number of data points.
    xdata : array-like
        A 1D array of independent variable values of shape (M,).
    ydata : array-like
        A 2D array of observed dependent variable values of shape (N, M), where each row corresponds
        to a dependent variable.
    p0 : array-like, optional
        Initial guess for the parameters to be optimized.
        Only used in the first minimization run (thereafter random).
    n_minimize_runs : int, optional
        Number of local optimization (minimization) runs to perform, each with a different
        starting point (first uses `p0`, others use DE output).
    n_de_runs : int, optional
        Number of differential evolution runs per minimization run to generate good starting
        guesses; the best DE result is used to initialize the local minimization.
    minimize_kwargs : dict
        Additional keyword arguments passed to `scipy.optimize.minimize`.
    differential_evolution_kwargs : dict
        Additional keyword arguments passed to `scipy.optimize.differential_evolution_kwargs`.
        
    Returns
    -------
    p_opt : ndarray
        Optimized parameter values that best fit the model.
    score : float
        The mean R² score achieved by the optimized parameters.
    success : bool
        Whether the finally used optimization result was flagged as successful.
        
    Notes
    -----
    - This function uses `scipy.optimize.minimize` and `scipy.optimize.differential_evolution` for optimization.
    - R² scores are computed between the predicted and observed dependent variable arrays.
    - The loss minimized is <one minus the mean R² score>.
    
    See Also
        --------
        `scipy.optimize.minimize <https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.minimize.html>`_
        `scipy.optimize.differential_evolution <https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.differential_evolution.html>`_
        
    """
    
    implemented_fit_methods = ('mean r^2',)
    ydata_transpose = ydata.transpose()
    
    def load_get_obj_f(p):
        ypred = f(xdata, p).transpose()
        return 1 - r2_score(ypred, ydata_transpose,
                          multioutput=r2_score_multioutput)
    
    best_result = None
    for i in range(n_minimize_runs):
        if show_progress:
            print(f'\n\nMinimization run {i+1}:')
            print('------------------\n')
        
        best_de_result = None
        
        bounds_de = differential_evolution_kwargs['bounds']
        
        if i>0:
            # p0 = np.random.rand(1)*random_param_bound*np.random.rand(*p0.shape)
            for j in range(n_de_runs):
                bounds_de_to_use = []
                for (b1, b2) in bounds_de:
                    bounds_de_to_use.append((b1, b2/(10**j)))
                differential_evolution_kwargs_to_use = differential_evolution_kwargs.copy()
                differential_evolution_kwargs_to_use['bounds'] = bounds_de_to_use
                if show_progress:
                    print(f'\n\tDifferential evolution run {i+1}.{j+1}:')
                    print('\t------------------------------')
                    print('\n\t\tRunning differential evolution (DE) to get the initial guess (x0) for this minimization run ...\n')
                    print('\t\t', differential_evolution_kwargs_to_use)
                result_de = differential_evolution(load_get_obj_f,
                                                   **differential_evolution_kwargs_to_use)
                
                if show_progress:
                    print('\n\t\tDE run complete.')
                    print('\t\tres.x =', result_de.x)
                    print('\t\tR^2 =', 1. - result_de.fun)
                    print('\t\tSuccess =', result_de.success)
                
                if best_de_result is None or result_de.fun < best_de_result.fun:
                    best_de_result = result_de
        
        p0 = p0 if best_de_result is None else best_de_result.x
        best_result = best_de_result
        if show_progress:
            print('\n\tRunning minimization to get the final set of parameters ...')
            print(f'\tx0 = {p0}')
            if best_de_result is not None:
                print(f'\tInitial R^2 = {1. - best_de_result.fun}')
            print('\t\t', differential_evolution_kwargs)
        result = minimize(fun=load_get_obj_f,
                 x0=p0,
                 **minimize_kwargs)
        if show_progress:
            print('\n\tMinimization run complete.')
            print('\tres.x =', result.x)
            print('\tR^2 =', 1. - result.fun)
            print('\tSuccess =', result.success)
        if best_result is None or result.fun < best_result.fun:
            best_result = result
    
    load_get_obj_f(best_result.x)
    
    return best_result.x, 1. - best_result.fun, best_result.success
    


# -*- coding: utf-8 -*-
# NSKinetics: simulation of Non-Steady state enzyme Kinetics and inhibitory phenomena
# Copyright (C) 2025-, Sarang S. Bhagwat <sarangbhagwat.developer@gmail.com>
# 
# This module is under the MIT open-source license. See 
# https://github.com/sarangbhagwat/nskinetics/blob/main/LICENSE
# for license details.

import nskinetics as nsk
import numpy as np
from warnings import filterwarnings

def test_simple_ESP_inhib_predefined(plot=True, 
                                     show_progress=False,
                                     show_output=True,
                                     show_warnings=True):
    if not show_warnings: filterwarnings('ignore')
    
    kcat = 32.
    kon = 1e2
    koff = 1e1
    
    # Default parameters for basic examples
    
    # Initial concentrations
    E_conc = 1e-3
    S_conc = 1e-1
    I_CI_conc = 1e-3
    I_NCI_conc = 0. #!!! debug NCI solve_ivp; happens with rxn strings too
    I_UCI_conc = 1e-3
    I_MBI_conc = 1e-3
    
    # Inhibition kinetics parameters
    kon_CI = 100
    koff_CI = 50
    kcat_CI = 30
    
    kon_ei=100
    koff_ei=90
    kon_es_esi_NCI=80
    koff_es_esi_NCI=70
    kon_ei_esi=75
    koff_ei_esi=60
    
    kon_es_esi_UCI = 60
    koff_es_esi_UCI = 50

    kon_MBI = 50
    koff_MBI = 20
    kstabilize_MBI = 40
    
    sp_sys = nsk.SpeciesSystem('multipurpose_sp_sys', ['E', 'S', 'ES', 'P', 
                                                   'I_CI', 'EI_CI', 'Q',
                                                   'I_NCI', 'EI_NCI', 'ESI_NCI',
                                                   'I_UCI', 'ESI_UCI',
                                                   'I_MBI', 'EI_MBI_unstable', 'EI_MBI_stable',
                                                   
                                                   ],
                           concentrations=[E_conc, S_conc, 0., 0., 
                                           I_CI_conc, 0., 0.,
                                           I_NCI_conc, 0., 0.,
                                           I_UCI_conc, 0.,
                                           I_MBI_conc, 0., 0.,
                                           ])
    
    MM_rxns = nsk.EnzymeSubstrateProduct(ID='MM', enzyme='E', substrate='S', es_complex='ES', product='P',
                              kon=kon, koff=koff, kcat=kcat, 
                              species_system=sp_sys)
    
    CI_rxns = nsk.CompetitiveInhibition(ID='CI', enzyme='E', substrate='I_CI', es_complex='EI_CI', product='Q',
                                    kon=kon_CI, koff=koff_CI, kcat=kcat_CI, 
                                    species_system=sp_sys)
    
    NCI_rxns = nsk.NonCompetitiveInhibition(ID='NCI', enzyme='E', inhibitor='I_NCI',
                                        substrate='S',
                                        ei_complex='EI_NCI',
                                        es_complex='ES',
                                        esi_complex='ESI_NCI',
                                        kon_ei=kon_ei, koff_ei=koff_ei,
                                        kon_es_esi=kon_es_esi_NCI, koff_es_esi=koff_es_esi_NCI,
                                        kon_ei_esi=kon_ei_esi, koff_ei_esi=koff_ei_esi,
                                        species_system=sp_sys)
    
    UCI_rxns = nsk.UnCompetitiveInhibition(ID='UCI', inhibitor='I_UCI', es_complex='ES',
                                       esi_complex='ESI_UCI',
                                       kon_es_esi=kon_es_esi_UCI, koff_es_esi=koff_es_esi_UCI, 
                                       species_system=sp_sys)
    
    MBI_rxns = nsk.MechanismBasedInhibition(ID='MBI', enzyme='E', inhibitor='I_MBI', 
                                            ei_unstable_complex='EI_MBI_unstable', 
                                            ei_stable_complex='EI_MBI_stable', 
                                            kon=kon_MBI, koff=koff_MBI, 
                                            kstabilize=kstabilize_MBI, 
                                            species_system=sp_sys)
    
    rxn_sys = nsk.RxnSys(ID='rxn_sys', 
                           reactions=[
                                      MM_rxns, 
                                      CI_rxns, 
                                      NCI_rxns, 
                                      UCI_rxns,
                                      MBI_rxns,
                                      ], 
                           species_system=sp_sys)
    

    # print(rxn_sys)
    # breakpoint()
    # Simulate the ReactionSystem
    
    rxn_sys.solve(t_span=[0, 10000],
                  sp_conc_for_events={'S':1e-6},)
    
    # Plot results
    if plot:
        rxn_sys.plot_solution()
    
    # Test absolute results
    assert np.allclose(rxn_sys._solution['t_events'], 
                       np.array([8382.63919662]), 
                       rtol=1e-3, atol=1e-3)
    
    assert np.allclose(rxn_sys._solution['y_events'], 
                       np.array([[4.82309338e-05],
                                 [1.00000000e-06],
                                 [4.38485153e-10],
                                 [1.00071398e-01],
                                 [8.30604981e-05],
                                 [7.85510180e-09],
                                 [9.17764773e-04],
                                 [1.00000000e-20],
                                 [1.00000000e-20],
                                 [1.00000000e-20],
                                 [9.99999745e-04],
                                 [5.26187882e-13],
                                 [4.87484236e-05],
                                 [5.59811388e-09],
                                 [9.52139030e-04]]), 
                       rtol=1e-3, atol=1e-8)
    
    # Make the same rxn_sys from strings
    
    reactions = [
                f'E + S <-> ES; kf={kon}, kb = {koff}',
                f'ES -> E + P; kf = {kcat}',
                
                f'E + I_CI <-> EI_CI; kf={kon_CI}, kb={koff_CI}',
                f'EI_CI -> E + Q; kf={kcat_CI}',
                
                f'E + I_NCI <-> EI_NCI; kf={kon_ei}, kb={koff_ei}',
                f'S + EI_NCI <-> ESI_NCI; kf={kon_ei_esi}, kb={koff_ei_esi}',
                f'ES + I_NCI <-> ESI_NCI; kf={kon_es_esi_NCI}, kb={koff_es_esi_NCI}',
                
                f'ES + I_UCI <-> ESI_UCI; kf={kon_es_esi_UCI}, kb={koff_es_esi_UCI}',
                
                f'E + I_MBI <-> EI_MBI_unstable; kf={kon_MBI}, kb={koff_MBI}',
                f'EI_MBI_unstable -> EI_MBI_stable; kf = {kstabilize_MBI}'
                ]
    
    # Generate a ReactionSystem from strings
    rxn_sys_from_str = nsk.ReactionSystem(ID='rxn_sys_from_str', 
                                     reactions=reactions,
                                     species_system=sp_sys)
    
    # print(rxn_sys_from_str)
    
    # Simulate
    sp_sys.concentrations = np.array([E_conc, S_conc, 0., 0., 
                    I_CI_conc, 0., 0.,
                    I_NCI_conc, 0., 0.,
                    I_UCI_conc, 0.,
                    I_MBI_conc, 0., 0.,
                    ])
    
    rxn_sys_from_str.solve(t_span=[0, 10000],
                  sp_conc_for_events={'S':1e-6},)
    
    # Plot results
    if plot:
        rxn_sys_from_str.plot_solution()
        
    # Test comparative results

    assert np.allclose(rxn_sys._solution['t_events'], 
                       rxn_sys_from_str._solution['t_events'], 
                       rtol=1e-3, atol=1e-3)
    
    assert np.allclose(rxn_sys._solution['y_events'], 
                       rxn_sys_from_str._solution['y_events'], 
                       rtol=1e-3, atol=1e-8)
    
    # Test comparative reaction strings
    for (r1, r2) in zip(rxn_sys.reactions_flattened, rxn_sys_from_str.reactions_flattened):
        r1_str = r1.__str__()
        r2_str = r2.__str__()
        assert r1_str[r1_str.index(':')+1:] == r2_str[r2_str.index(':')+1:]
        
    return rxn_sys